/*
 * edit.c
 * line editing functions for cancan
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifndef STDASCII
#include <ctype.h>
#endif
#include <unistd.h>

#include "cancan.h"
#include "cmd.h"
#include "term.h"

int strncasecmp();
char *strdup();

static void end_of_line(), beginning_of_line(), redraw_input_line();
static void del_char_right(), del_char_left(), transpose_chars();
static void enter_into_history(), enter_line(), kill_to_eol();
static void next_word(), prev_word(), del_word_right(), del_word_left();
static void prev_line(), next_line(), prev_char(), next_char();
static void complete_word(), complete_line();

static void put_in_history();
static void swap_history();
static void overtype();

/* completion list */
static char *words[WORDSKEPT];
static int wordindex;

/* history buffer */
char *hist[MAXHIST];	/* saved history lines */
int curline;		/* current history line */
int pickline;		/* line to pick history from */

struct defbind_node defbinds[] = {
  { "enter",	"\n",		NULL,			enter_line },
  { "return",	"\r",		NULL,			enter_line },
  { "bs",	"\b",		"&del-char-left",	del_char_left },
  { "del",	"\177",		"&del-char-left",	del_char_left },
  { "tab",	"\t",		"&complete-word",	complete_word },
  { "left",	key_left,	"&prev-char",		prev_char },
  { "right",	key_right,	"&next-char",		next_char },
  { "up",	key_up,		"&prev-line",		prev_line },
  { "down",	key_down,	"&next-line",		next_line },
  { "C-a",	"\001",		"&beginning-of-line",	beginning_of_line },
  { "C-b",	"\002",		"&prev-char",		prev_char },
  { "C-c",	"\003",		NULL,			intrsig },
  { "C-d",	"\004",		"&del-char-left",	del_char_right },
  { "C-e",	"\005",		"&end-of-line",		end_of_line },
  { "C-f",	"\006",		"&next-char",		next_char },
  { "C-k",	"\013",		"&kill-to-eol",		kill_to_eol },
  { "C-l",	"\014",		"&redraw-input-line",	redraw_input_line },
  { "C-n",	"\016",		"&next-line",		next_line },
  { "C-p",	"\020",		"&prev-line",		prev_line },
  { "C-t",	"\024",		"&transpose-chars",	transpose_chars },
  { "C-y",	"\031",		"&enter-into-history",	enter_into_history },
  { "C-z",	"\032",		NULL,			suspend_cancan },
  { "M-tab",	"\033\t",	"&complete-line",	complete_line },
  { "M-b",	"\033b",	"&prev-word",		prev_word },
  { "M-d",	"\033d",	"&del-word-right",	del_word_right },
  { "M-f",	"\033f",	"&next-word",		next_word },
  { "M-del",	"\033\177",	"&del-word-left",	del_word_left },
  { "",		NULL,		NULL,			NULL }
};

/*
 * redraw the input line
 */
static void redraw_input_line()
{
    clearinpline();
    redraw_input();
}

/*
 * transpose two characters to the left
 */
static void transpose_chars()
{
    int k;
    unsigned char c;
    if((k = pos)) {
	if(k-- >= edlen && k) k--;
	moveto(k);
	c = edbuf[k];
	overtype(edbuf[k + 1]);
	overtype(c);
    }
}

/*
 * erase everything to the end of line
 */
static void kill_to_eol()
{
    int cline = CURLINE(pos);
    fputs(clreoln, stdout);
    if(CURLINE(edlen) > cline) {
	printf("\n%s", clreoscr);
	optmove(cline + 1, 0, cline, CURCOL(pos));
    }
    edbuf[pos] = '\0';
    edlen = pos;
}

/*
 * move cursor to end of line
 */
static void end_of_line()
{
    moveto(edlen);
}

/*
 * move cursor to beginning of line
 */
static void beginning_of_line()
{
    moveto(0);
}

/*
 * delete a character to the right
 */
static void del_char_right()
{
    if(pos < edlen)
	deletechars(1);
}

/*
 * delete a character to the left
 */
static void del_char_left()
{
    if(pos) {
	moveto(pos - 1);
	deletechars(1);
    }
}

/*
 * enter a line into history, but don't do anything else
 */
static void enter_into_history()
{
    clearinpline();
    put_in_history(edbuf);
    pickline = curline;
    strcpy(edbuf, prefixstr);
    pos = edlen = strlen(edbuf);
}

/*
 * enter a line
 */
static void enter_line()
{
    char *p;
    int ls = (pos >= edlen) ? 1 : 1 + CURLINE(edlen) - CURLINE(pos);
    do newline(); while(--ls); 
    if(capfile) fprintf(capfile, "%s\n", edbuf);
#ifdef NEWBIE
    denewbiefy();
#endif
    inserted_next[0] = '\0';
    if(verbatim && edbuf[0] != '#') {
	edbuf[edlen] = '\n';
	send_to_host(edbuf, edlen + 1);
	edbuf[edlen] = '\0';
    } else {
	if(edattr) { edattr = 0; fputs(edattrend, stdout); }
	new_settings = 0;
        verbose = 1;
	parse(edbuf, 1, NULL);
	verbose = 0;
	if(new_settings)	/* user defined something -- save */
	    save_settings();
    }
    check_prompt();
    p = hist[curline ? curline - 1 : MAXHIST - 1];
    /* don't put identical lines in history, nor short ones */
    if(edlen > 1 && (!p || strcmp(edbuf, p)))
	put_in_history(edbuf);
    pickline = curline;
    strcpy(edbuf, inserted_next[0] ? inserted_next : prefixstr);
    pos = edlen = strlen(edbuf);
    redraw_input();
}

/*
 * move one word forward
 */
static void next_word()
{
    int i;
    for(i = pos; !islabel(edbuf[i]) && edbuf[i]; i++);
    while(islabel(edbuf[i])) i++;
    moveto(i);
}

/*
 * move one word backward
 */
static void prev_word()
{
    int i;
    for(i = pos; i && !islabel(edbuf[i - 1]); i--);
    while(i && islabel(edbuf[i - 1])) i--;
    moveto(i);
}

/*
 * delete word to the right
 */
static void del_word_right()
{
    int i;
    for(i = pos; !islabel(edbuf[i]) && i < edlen; i++);
    while(islabel(edbuf[i])) i++;
    deletechars(i - pos);
}

/*
 * delete word to the left
 */
static void del_word_left()
{
    int i, n;
    for(i = pos; i && !islabel(edbuf[i - 1]); i--);
    while(i && islabel(edbuf[i - 1])) i--;
    n = pos - i;
    moveto(i);
    deletechars(n);
}

/*
 * get previous line from history list
 */
static void prev_line()
{
    int i = pickline - 1;
    if(i < 0) i = MAXHIST - 1;
    swap_history(i);
}

/*
 * get next line from history list
 */
static void next_line()
{
    int i = pickline + 1;
    if(i == MAXHIST) i = 0;
    swap_history(i);
}

/*
 * move one char backward
 */
static void prev_char()
{
    if(pos)
	moveto(pos - 1);
}

/*
 * move one char forward
 */
static void next_char()
{
    if(pos < edlen)
	moveto(pos + 1);
}

/*
 * execute string as if typed
 */
void key_run_command(cmd)
char *cmd;
{
    clearinpline();
    if(!edattr) fputs(edattrbeg, stdout);
    printf("%s%s\n", cmd, edattrend);  edattr = 0;
    if(capfile) fprintf(capfile, "%s\n", cmd);
    parse(cmd, 1, NULL);
    check_prompt();
    redraw_input();
}

/*
 * match and complete a word referring to the word list
 */
static void complete_word()
{
    int i, j;
    if(pos) {
	/* find word start */
	if((j = pos)) while(!strchr(DELIM, edbuf[j - 1]) && --j);
	if(j == pos) return;
     
	if(edbuf[j] == '#' && (i = pos - j - 1)) {
	    /* try to complete built-in command */
	    cmdstruct *p;
	    for(p = commands; p->name; p++)
		if(p->funct && !strncmp(p->name, edbuf + j + 1, i))
		    insertchars(p->name + i, strlen(p->name + i));
	    return;
	}
	/* scan word list for match */
	i = wordindex;
	do {
	    if(--i < 0) i = WORDSKEPT - 1;
	    if(!words[i]) break;
	    if(!strncasecmp(words[i], edbuf + j, pos - j)) {
		insertchars(words[i] + pos - j, strlen(words[i] + pos - j));
		break;
	    }
	} while(i != wordindex);
    }
}

/*
 * match and complete entire lines backwards in history
 */
static void complete_line()
{
    if(edlen) {
	int i = curline;
	while(--i != curline) {
	    if(i < 0) i = MAXHIST - 1;
	    if(hist[i] && !strncmp(hist[i], edbuf, edlen)) {
		clearinpline();
		strcpy(edbuf, hist[i]);
		pos = edlen = strlen(edbuf);
		redraw_input();
		break;
	    }
	}
    }
}

/*
 * put string in history at current position
 * (string is assumed to be trashable)
 */
static void put_in_history(str)
char *str;
{
    char *p;
    if(hist[curline]) free(hist[curline]);
    hist[curline] = strdup(str);
    if(++curline == MAXHIST) curline = 0;
  
    /* split into words and put into completion list */
    for(p = strtok(str, DELIM); p; p = strtok(NULL, DELIM)) {
	if(strlen(p) >= 3 && p[0] != '#')	/* no commands/short words */
	    put_word(p);
    }
}

/*
 * swap editline with a history line
 */
static void swap_history(line)
int line;
{
    if(hist[line]) {
	if(hist[pickline]) free(hist[pickline]);
	hist[pickline] = strdup(edbuf);
	pickline = line;
	clearinpline();
	strcpy(edbuf, hist[line]);
	pos = edlen = strlen(edbuf);
	redraw_input();
    }
}

/*
 * put word in word completion list
 */
void put_word(word)
char *word;
{
    if(words[wordindex]) free(words[wordindex]);
    words[wordindex] = strdup(word);
    if(++wordindex == WORDSKEPT) wordindex = 0;
}

/*
 * print a char on current position (overwrite), advance position
 */
static void overtype(c)
char c;
{
    edbuf[pos++] = c;
    putchar(c);
    if(!CURCOL(pos))
	putchar('\n');
}
